import numpy as np
from json import JSONEncoder

class NumpyArrayEncoder(JSONEncoder):
    """
    Encode numpy to dump in json
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)